import _sortedIndex from "lodash-es/sortedIndex";
import _sortBy from "lodash-es/sortBy";
import _reverse from "lodash-es/reverse";
import _range from "lodash-es/range";
import _merge from "lodash-es/merge";
import _map from "lodash-es/map";
import _isUndefined from "lodash-es/isUndefined";
import _isNumber from "lodash-es/isNumber";
import _isFunction from "lodash-es/isFunction";
import _includes from "lodash-es/includes";
import _fromPairs from "lodash-es/fromPairs";
import _forEach from "lodash-es/forEach";
import _flatMap from "lodash-es/flatMap";
import _filter from "lodash-es/filter";
import _entries from "lodash-es/entries";
import { assert } from '../assert';
import { Options, builder, SYMBOL_STATE, SYMBOL_LOG } from '@escapace/fluent';
import { SYMBOL_ANIMATION } from '../types';
import { TypeAction } from './types';
import { TypeAction as ListTypeAction, TypeKeyframes } from '../keyframe-list/types';
import { simple } from '../simple';
import lerpArray from 'lerp-array'; // eslint-disable-next-line @typescript-eslint/no-explicit-any

export const reducer = log => {
  return {
    type: SYMBOL_ANIMATION,
    isEmpty: log.length === 0
  };
};

const animationData = ( // eslint-disable-next-line @typescript-eslint/no-explicit-any
log, options) => {
  const keyMap = new Map();
  const listNames = new Set();

  const _log = _flatMap(_reverse(_filter(log, ({
    type
  }) => type === TypeAction.KeyframeList)), (listAction, listIndex) => {
    const listName = listAction.payload[SYMBOL_STATE].name;

    if (listNames.has(listName)) {
      throw new Error('TODO:');
    } else {
      listNames.add(listName);
    }

    return _flatMap(_reverse(_filter(listAction.payload[SYMBOL_LOG], ({
      type
    }) => type === ListTypeAction.Keyframes)), (propertyAction, propertyIndex) => {
      const propertyName = propertyAction.payload.name;
      const propertyType = propertyAction.payload.type;
      const propertyInterpolation = _isFunction(propertyAction.payload.interpolation) ? propertyAction.payload.interpolation : _includes([TypeKeyframes.Number, TypeKeyframes.NumberArray], propertyType) ? lerpArray : x => x;
      const keyData = {
        listName,
        propertyName,
        listIndex,
        propertyIndex,
        propertyType,
        propertyInterpolation
      };
      const key = `${listName}${propertyName}`;
      keyMap.set(key, keyData);
      const keyframes = propertyAction.payload.keyframes;
      return {
        keyData,
        keyframes,
        key
      };
    });
  });

  const data = _fromPairs(_map(options.iterations, iteration => {
    const properties = new Map();

    _forEach(_log, ({
      keyframes,
      key
    }) => {
      _forEach(keyframes, (frame, progress) => {
        let indice;

        if (properties.has(key)) {
          indice = properties.get(key);
        } else {
          properties.set(key, new Map());
          indice = properties.get(key);
        }

        const kf = _isFunction(frame) ? frame(iteration) : frame; // eslint-disable-next-line @typescript-eslint/no-non-null-assertion

        indice.set(Number.parseFloat(progress), kf);
      });
    });

    const mappedProperties = new Map();
    properties.forEach((value, key) => {
      const arr = Array.from(value.entries());
      mappedProperties.set(keyMap.get(key), _sortBy(arr, [value => value[0]]));
    });
    return [iteration, Array.from(mappedProperties.entries())];
  }));

  return data;
}; // eslint-disable-next-line @typescript-eslint/no-explicit-any


const keyframeProduct = (keyMap, value) => {
  return {
    [keyMap.listName]: {
      [keyMap.propertyName]: value
    }
  };
};

export const animation = /*#__PURE__*/builder([{
  [Options.Type]: TypeAction.KeyframeList,
  [Options.Once]: false,
  [Options.Keys]: ['add'],
  [Options.Reducer]: reducer,
  [Options.Interface]: dispatch => ({
    add(value) {
      assert.list(value); // TODO: duplicate name test

      return dispatch({
        type: TypeAction.KeyframeList,
        payload: simple(value)
      });
    }

  })
}, {
  [Options.Type]: TypeAction.Reduce,
  [Options.Once]: false,
  [Options.Dependencies]: [TypeAction.KeyframeList],
  [Options.Keys]: ['reduce'],
  [Options.Reducer]: reducer,
  [Options.Interface]: (_, log) => ({
    reduce(options) {
      let iterations;

      if (_isUndefined(options === null || options === void 0 ? void 0 : options.iterations)) {
        iterations = [0];
      } else {
        assert.iterations(options === null || options === void 0 ? void 0 : options.iterations);
        iterations = _isNumber(options === null || options === void 0 ? void 0 : options.iterations) ? _range(0, options === null || options === void 0 ? void 0 : options.iterations) : options === null || options === void 0 ? void 0 : options.iterations;
      }

      const data = animationData(log, {
        iterations
      }); // listNames.
      // console.log()

      return (progress = 0, iteration = 0) => {
        if (progress > 100) {
          progress = 100;
        } // console.log(data[0])


        const object = _merge({}, ..._map(data[iteration], ([keyMap, encounters]) => {
          const interpolation = keyMap.propertyInterpolation;

          if (encounters.length === 1 || progress === 0) {
            return keyframeProduct(keyMap, interpolation(encounters[0][1].value, encounters[0][1].value, 0));
          }

          const index = _sortedIndex(_map(encounters, enc => enc[0]), progress);

          if (_isUndefined(encounters[index])) {
            return keyframeProduct(keyMap, interpolation(encounters[index - 1][1].value, encounters[index - 1][1].value, 0));
          }

          const [nextIndex, nextContent] = encounters[index];
          const [previousIndex, previousContent] = encounters[index - 1];

          if (_includes([TypeKeyframes.Number, TypeKeyframes.NumberArray], keyMap.propertyType)) {
            const alpha = (progress - previousIndex) / (nextIndex - previousIndex);
            return keyframeProduct(keyMap, interpolation(previousContent.value, nextContent.value, _isFunction(previousContent.easing) ? previousContent.easing(alpha) : alpha));
          } else {
            return keyframeProduct(keyMap, previousContent.value);
          }
        }));

        return Object.assign(object, {
          [Symbol.iterator]: () => _entries(object)[Symbol.iterator]()
        });
      };
    }

  })
}]); // import { list } from '../keyframe-list/domain-language'
// const qwe = animation()
//   .add(
//     list()
//       .name('listA')
//       .number('listA0', {
//         0: () => ({ value: 1 }),
//         100: { value: 30 }
//       })
//       .number('listA1', {
//         0: () => ({ value: 1 }),
//         100: { value: 50 }
//       })
//       .string('listA2', {
//         0: () => ({ value: 'qwe' }),
//         100: () => ({ value: 'abcd' })
//       })
//   )
//   .add(
//     list()
//       .name('listB')
//       .number('listB0', {
//         0: () => ({ value: [0, 0], easing: 'easeInSine()' }),
//         100: () => ({ value: [79, 12] })
//       })
//       .boolean('listB1', {
//         0: { value: true },
//         100: { value: false }
//       })
//   )
//   .reduce()
// const iter = Array.from(qwe())
// console.log(iter)
//# sourceMappingURL=domain-language.js.map
